#
# ASAP Answer Set Application Programming - ASAP Python API
#

#Copyright (c) 2015, Peter Schueller <peter.schuller@marmara.edu.tr>
#Copyright (c) 2015, Antonius Weinzierl <weinzierl@kr.tuwien.ac.at>
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
#1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
#2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
#3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import sys
import re
import logging

def debug(msg):
  # TODO replace below
  logging.debug(msg)

def warn(msg):
  # TODO replace below
  logging.warning(msg)

rCident = re.compile(r'[a-z_][A-Z0-9a-z_]*')

class Term:
  '''
  a term or atom in an interpretation
  '''
  # TODO maybe pull out of class and make module function
  @staticmethod
  def fromNestedTuple(nestedtuple):
    '''
    instantiate a term from a tuples containing tuples/strings/integers in any nesting level
    '''
    if not isinstance(nestedtuple,tuple) or len(nestedtuple) == 0:
      raise Exception("fromNestedTuple requires nonempty tuple as input (got {})".format(repr(nestedtuple)))
    if not isinstance(nestedtuple[0], (str,int)):
      raise Exception("fromNestedTuple requires int or string as first tuple element (got {})".format(repr(nestedtuple[0])))
    args = []
    for a in nestedtuple[1:]:
      if isinstance(a,int) or isinstance(a,str):
        args.append(Term(a))
      else:
        args.append(Term.fromNestedTuple(a))
    return Term(nestedtuple[0], tuple(args))
  def __init__(self,pred,args=[]):
    '''
    initialize term with predicate (str or int) and iterable arguments
    each argument is a Term
    '''
    if not isinstance(pred,str) and not isinstance(pred,int):
      raise Exception("Term must have str or int predicate (is '{}')".format(repr(pred)))
    if not isinstance(args,(list,tuple)):
      raise Exception("Term must have list or tuple as arguments (has '{}')".format(repr(args)))
    if isinstance(pred,int) and len(args) > 0:
      raise Exception("Term(int) '{}' cannot have arguments".format(pred))
    if isinstance(pred,str) and not ((pred[0] == '"') ^ (pred[-1] != '"')):
      raise Exception("Term(str) '{}' needs either opening AND closing quotes, or no quotes at all".format(pred))
    if isinstance(pred,str) and pred[0] != '"' and not rCident.match(pred):
      raise Exception("Term(str) '{}' must be CIDENT".format(pred))
    if not all([ (isinstance(arg,str) or isinstance(arg,int) or isinstance(arg,Term)) for arg in args]):
      raise Exception("Term must have Term arguments (has '{}')".format(','.join([repr(x) for x in args])))
    self.tupl = tuple([pred]+list(args))
  @property
  def pred(self):
    return self.tupl[0]
  @property
  def args(self):
    return self.tupl[1:]
  def as_tuple(self):
    # return internal tuple, it is read-only (and contains only Terms which are read-only)
    return self.tupl
  def __int__(self):
    if not isinstance(self.tupl[0],int):
      raise Exception("requested integer of api.Term '{}' which is not of type integer".format(pred))
    return self.tupl[0]
  def unquoted(self):
    # returns predicate constant or predicate string without quotes
    assert(len(self.tupl) == 1 and isinstance(self.pred,str))
    if self.pred[0] == '"' and self.pred[-1] == '"':
      return self.pred[1:-1]
    else:
      return self.pred
  def __hash__(self):
    return hash(self.tupl)
  def __str__(self):
    if len(self.tupl) == 1:
      return str(self.tupl[0])
    else:
      return '{}({})'.format(self.tupl[0], ','.join([str(x) for x in self.tupl[1:]]))
  def __repr__(self):
    if len(self.tupl) == 1:
      return 'Term({})'.format(self.tupl[0])
    else:
      return 'Term({}({}))'.format(self.tupl[0], ','.join([repr(x) for x in self.tupl[1:]]))

class Event:
  '''
  Event representation in ASAP
  '''
  firstEventTimestamp = time.time()
  def __init__(self, term):
    self.term = term
    self.timestamp = time.time() - Event.firstEventTimestamp
    #debug("constructed event "+str(self))
  def __str__(self):
    return "Event({},{})".format(self.term, self.timestamp)
  def __repr__(self):
    return self.__str__()
  def __eq__(self, other):
    return self.term == other.term

class EventQueue:
  '''
  ASAP realizes this API and gives it to the EventProvider for providing events
  (implementation in derived class)
  '''
  def append(self, event):
    '''
    add event to end of queue
    '''
    debug('enqueued event '+str(event))
  def immediate(self, event):
    '''
    add (immediate) event to beginning of queue
    '''
    debug('enqueued immediate '+str(event))

  def pop(self):
    '''get/pop an item from front'''
    pass

  def __iter__(self):
    '''
    provide iterator over queue contents
    '''
    return iter([])

class EventAPI:
  '''
  plugins realizing this API can add events to the event queue
  '''
  def __init__(self):
    self.queue = None
  def setQueue(self, queue):
    '''
    tell the EventAPI which EventQueue should be used
    '''
    assert(self.queue is None)
    self.queue = queue

class Environment:
  '''
  base class for environments
  an environment represents the view of the external world by the program
  applications will need to define its own environment
  TODO or could plugins define the environment and we automatically mix them together?
  '''
  def __str__(self):
    return 'Environment()'

class Input:
  '''
  base class for providing input to actions and externals
  '''
  def __init__(self, environment, inputs, interpretation=None):
    """
    environment: derived from api.Environment
    inputs: tuple of api.Term
    interpretation: set of api.Term
    """
    self.environment = environment
    self.inputs = inputs
    self.interpretation = interpretation
  def __str__(self):
    interpretationinfo = 'empty'
    if self.interpretation is not None:
      interpretationinfo = 'contains {} terms'.format(len(self.interpretation))
    return "Input(environment={},inputs={},interpretation {})".format(
      self.environment, self.inputs, interpretationinfo)
  def __repr__(self):
    return self.__str__()

class ExternalInput(Input):
  '''
  provides all data that an external receives for its computation
  '''
  def __init__(self, environment, inputs, interpretation=None):
    Input.__init__(self, environment, inputs, interpretation)

class External:
  '''
  a handler for an external
  '''
  @property
  def interface(self):
    '''
    return interface of the external as a tuple
    (name,in_arity,in_arity_mandatory,out_arity)
    e.g. Xt.temperature[sensorname](reading) would have interface
      ('temperature',1,1,1)
    e.g. Xt.concat[str1,str2,...](out) would have interface
      ('concat',99,2,1)
    '''
    pass
  def query(self, einput):
    '''
    evaluate the external relative to the ExternalInput einput
    return an iterable of tuples of output terms (api.Term objects) for this external
    (if the arity is 1, the output shall be an iterable over tuples of size one)
    '''
    return []

class ExternalAPI:
  '''
  plugins realizing this API can provide externals
  '''
  def externals(self):
    '''
    provide iterable of External objects
    '''
    return []

class ActionInput(Input):
  '''
  provides all data that an action receives for its execution
  '''
  def __init__(self, environment, inputs, interpretation=None):
    Input.__init__(self, environment, inputs, interpretation)

class Action:
  '''
  a handler for an action
  '''
  @property
  def interface(self):
    '''
    return interface of the action as a tuple
    (name,in_arity,in_arity_mandatory)
    e.g. Do.drawString(align,pos,string[,color]) would have interface
      ('drawString',4,3)
    '''
    pass
  def execute(self, ainput):
    '''
    execute the action relative to the ActionInput ainput
    '''
    pass

class DoAPI:
  '''
  plugins realizing this API can provide actions
  '''
  def actions(self):
    '''
    provide iterable of Action objects
    '''
    return []
  def beforeExecution(self):
    '''
    this method is called before executing actions from an answer set (even if no actions are executed)
    '''
    pass
  def afterExecution(self):
    '''
    this method is called after executing actions from an answer set (even if no actions are executed)
    '''
    pass

class Backend:
  '''
  a backend/engine realizes this interface
  '''
  def evaluate(self, program, plugins=[], environment=None, fluentfile=None):
    '''
    evaluate given ASAP program
    program is the program to evaluate 
    plugins is an iterable of objects of type
      DoAPI, ExternalAPI, EventAPI, or combinations thereof
    environment is the initial environment
    fluentfile is the file to load fluents from (json format)
    '''
    pass

# TODO provide base class for plugin?
# TODO provide method for registering plugins



from . import api

import os
import sys
import threading
import subprocess
import logging

FAIL_ON_MALFORMED_ACTIONS=True

def debug(msg):
  logging.debug(msg)

def warn(msg):
  logging.warning(msg)

class EventQueueImpl(api.EventQueue):
  def __init__(self):
    self.elements = []
    self.cv = threading.Condition()

  def append(self, event):
    try:
      self.cv.acquire()
      self.elements.append(event)
      api.EventQueue.append(self, event)
      self.cv.notify()
    finally:
      self.cv.release()

  def immediate(self, event):
    try:
      self.cv.acquire()
      self.elements.insert(0,event)
      api.EventQueue.immediate(self, event)
      self.cv.notify()
    finally:
      self.cv.release()

  def pop(self):
    try:
      self.cv.acquire()
      while len(self.elements) == 0:
        self.cv.wait()
      ret = self.elements.pop(0)
    finally:
      self.cv.release()
    return ret

  def __iter__(self):
    return iter(self.elements)

class FluentStorage(dict):
  def __init__(self, pairs=None):
    if pairs:
      assert(all([isinstance(p[0], api.Term) and isinstance(p[1], api.Term) for p in pairs]))
      dict.__init__(self, pairs)
    else:
      dict.__init__(self)
  def __getitem__(self,key):
    if not isinstance(key, api.Term):
      raise Exception("attempted to get fluent '{}' which is not an api.Term".format(repr(key)))
    return dict.__getitem__(self, key)
  def __setitem__(self,key,value):
    if not isinstance(key, api.Term) or not isinstance(value, api.Term):
      raise Exception("attempted to set fluent '{}' to value '{}' where at least one is not an api.Term".format(repr(key), repr(value)))
    return dict.__setitem__(self, key)
  def nice(self):
    return '\t'+'\n\t'.join([
      '{} => {}'.format(str(x[0]), str(x[1]))
      for x in self.items()])

def extractFluents(answerset):
  '''
  extract FluentStorage from an iterable of Term
  fluents are represented as val(FluentName,Timepoint,FluentValue) where we extract timepoint 0
  '''
  assert(all([isinstance(x,api.Term) for x in answerset]))
  #debug("extracting fluents from answer set {}".format(repr(answerset)))
  keyValuePairsNext = [ (t.args[0],t.args[2]) for t in answerset if t.pred == 'val' and t.args[1].pred == 0 ]
  return FluentStorage(keyValuePairsNext)

def extractActions(answerset):
  '''
  extract sequence of actions from an iterable of Term
  actions are represented as action(<actionname>(<arguments>),<priority>)
  output of this function is a sequence of triples (<actionname>, <arguments>, <priority>)
  '''
  assert(all([isinstance(x,api.Term) for x in answerset]))
  actionTriples = [ (t.args[0].pred, t.args[0].args, t.args[1].pred) for t in answerset if t.pred == 'action' ]
  ret = sorted(actionTriples, key=lambda x: x[2])
  #debug('\n  '.join(["extracted action schedule:"] + map(lambda x: '{}/{}/{}'.format(*x), ret)))
  return ret

def malformedAction(name, args, msg):
  message = "malformed action '{}' with arguments '{}' ({})".format(name, args, msg)
  if FAIL_ON_MALFORMED_ACTIONS:
    raise Exception(message)
  else:
    warn(message + " (ignoring)")

def executeActionSchedule(doApis, actions, environment, answerset, actionSchedule):
  """
  doApis is obtained from asap.core.separateApis() [first element]
  actions is obtained from asap.core.actionsFromApis() [first element]
  environment is an environment
  answerset is the answer set
  actionSchedule is obtained via asap.core.extractActions()
  """
  for d in doApis:
    d.beforeExecution()
  for name, args, prio in actionSchedule:
    if name not in actions:
      debug("known actions: "+repr(actions.keys()))
      malformedAction(name, args, 'unknown action')
      continue
    a = actions[name]
    in_arity, in_arity_mandatory = a.interface[1:]
    if len(args) < in_arity_mandatory or len(args) > in_arity:
      malformedAction(name, args, 'wrong number of arguments')
      continue
    inp = api.ActionInput(environment, args, answerset)
    #debug("executing action '{}' with arguments '{}'".format(name, args))
    a.execute(inp)
  for d in doApis:
    d.afterExecution()

def separateApis(pl):
  logging.info("separateApis got {} plugins: {}".format(len(pl), repr(pl)))
  doApis = [x for x in pl if isinstance(x,api.DoAPI)]
  eventApis = [x for x in pl if isinstance(x,api.EventAPI)]
  externalApis = [x for x in pl if isinstance(x,api.ExternalAPI)]
  logging.info("separateApis found doApis {} eventApis {} externalApis {}".format(repr(doApis), repr(eventApis), repr(externalApis)))
  return doApis, eventApis, externalApis

def registerQueue(eventApis, eq):
  for api in eventApis:
    api.setQueue(eq)

def actionsFromApis(doApis):
  actions = {}
  for api in doApis:
    for action in api.actions():
      name = action.interface[0]
      if name in actions:
        raise Exception("duplicate action '{}' registered: {} vs. {}".format(name, repr(actions[name].interface), repr(action.interface)))
      actions[name] = action
  return actions

def externalsFromApis(externalApis):
  externals = {}
  for api in externalApis:
    for external in api.externals():
      name = external.interface[0]
      if name in externals:
        raise Exception("duplicate external '{}' registered: {} vs. {}".format(name, repr(externals[name].interface), repr(external.interface)))
      externals[name] = external
  return externals

def rewrite(program, target):
  '''
  rewrite ASAP program for one of the supported targets
  '''
  assert(target in ['hex', 'clasp'])
  rewriterpath = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'ASAP-Rewriter.jar')
  logging.info("rewriting program with target '{}':".format(target))
  if __debug__:
    logging.debug("ORIGINAL:")
    for o in program.split('\n'):
      logging.debug('OR '+o.strip('\n'))
  p = subprocess.Popen('java -jar {} -target={}'.format(rewriterpath, target),
    stdin=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
  out, err = p.communicate(bytes(program, 'utf-8'))
  out = out.decode('utf-8')
  if __debug__:
    logging.debug("REWRITTEN:")
    for o in out.split('\n'):
      logging.debug('RW '+o.strip('\n'))
  return out




from asap.core import api

import math

class DelayFromLevelExternal(api.External):
  '''
  compute milliseconds for stone drop depending on game level (difficulty)
  input: level (int)
  output: milliseconds (int)
  '''
  def __init__(self):
    # if easiness is 1, game becomes more difficult very fast
    # the higher this number, the longer the game will take
    self.easiness = 5.0 #10.0
  @property
  def interface(self):
    # one mandatory input, one output
    return ('delayFromLevel',1,1,1)
  def query(self, einput):
    assert(len(einput.inputs) == 1)
    level = int(einput.inputs[0])
    delay_ms = int(1000.0*math.exp(-level / self.easiness))
    return [ (api.Term(delay_ms),) ]

class Plugin(api.ExternalAPI):
  def externals(self):
    return [DelayFromLevelExternal()]





  def selectAnswerset(self, answersets):
    if len(answersets) == 0:
      raise RuntimeError("no answersets!")
    if len(answersets) != 1:
      if self.mode == 'acceptmultipleanswersets':
        return answersets[0]
      warn("more than one answerset!")
      asSetOfStrings = map(lambda a: set(map(str, a)), answersets)
      for idx, answerset in enumerate(asSetOfStrings):
        if idx == 0:
          continue
        for idxPrev, answersetPrev in enumerate(asSetOfStrings):
          if idxPrev >= idx:
            break
          plus = answerset - answersetPrev
          minus = answersetPrev - answerset
          #warn('plus {}'.format(plus))
          #warn('minus {}'.format(minus))
          diff = map(lambda p: p+'+', plus) + map(lambda m: m+'-', minus)
          warn("#{} vs #{}: {}".format(idx, idxPrev, ', '.join(sorted(diff))))
      for idx, answerset in enumerate(answersets):
        warnVerbose("Full Answer #{}: {}\n".format(idx, str(answerset)))
      raise RuntimeError("more than one answerset!")
    return answersets[0]

  def checkForASPErrors(self, answerset):
    # check for errors
    error = False
    for atom in answerset:
      atuple = getTupleValues(atom)
      if len(atuple) == 2 and atuple[0] == 'error':
        warn("ASP ERROR: {}".format(repr(atuple[1])))
        error = True
    if error:
      raise RuntimeError("encountered ASP error(s)!")

  def enqueueSoftEvents(self, actions):
    '''
    find soft events and execute them
    return unhandled actions
    '''
    ret = []
    for a in actions:
      aname, aargs, ser = a[0], a[1:-1], a[-1]
      if aname == 'softEvent':
        name = aargs[0]
        self.queue.put_front(MyEvent('softevent', name))
      else:
        ret.append(a)
    return ret




from asap.core import api

from tkinter import font
import tkinter as tk

import logging
import queue
import threading, sys, os, traceback

def debug(msg):
  #sys.stderr.write("D:"+msg+"\n")
  pass

def warn(msg):
  sys.stderr.write("W:"+msg+"\n")


class KeyboardHandler(api.EventAPI):
  def __init__(self, root):
    api.EventAPI.__init__(self)
    self.root = root

  def setQueue(self, queue):
    api.EventAPI.setQueue(self, queue)
    self.root.bind_all('<KeyPress>', self.keyPressEvent)

  def keyPressEvent(self, evt):
    if not self.queue:
      warn("tkinter.KeyboardHandler ignoring keyPressEvent because no queue registered")
      return
    debug('keyPressEvent {}/{}/{}'.format(evt.char, evt.keysym, evt.keycode))
    if evt.char == evt.keysym:
      # normal character
      self.queue.append(api.Event(api.Term.fromNestedTuple(('keyboard', ('normal', '"{}"'.format(evt.char))))))
    else:
      self.queue.append(api.Event(api.Term.fromNestedTuple(('keyboard', ('special', '"{}"'.format(evt.keysym))))))

class WindowHandler(api.DoAPI):
  def __init__(self, root):
    #api.DoAPI.__init__(self)

    self.root = root
    self.items = []
    self.drawq = queue.Queue()

    #self.root.geometry("800x600")
    self.root.geometry("1200x800")
    self.XMUL = 25
    self.YMUL = 25
    self.XOFFSET = self.XMUL
    self.YOFFSET = self.YMUL

    self.imageCache = {}

    self.main = tk.Frame(self.root)
    self.main.pack(fill=tk.BOTH, expand=tk.YES)
    self.canvas = tk.Canvas(self.main, bg="grey")
    self.canvas.pack(fill=tk.BOTH, expand=tk.YES)
    self.FONT = font.Font(family = 'Courier', size = self.YMUL-1)

    self.pixel_viewport = (1,1)
    self.updateViewport()

    self.my_actions = [
      DrawRectangleAction(self),
      DrawImageAction(self),
      DrawTextAction(self),
      DrawTextCenteredAction(self)]

    self.root.bind('<<DrawingUpdate>>', lambda x: self.drawingUpdate(x))

  def actions(self):
    return self.my_actions

  def beforeExecution(self):
    drawTextCentered = self.my_actions[3]
    drawTextCentered.resetCursor()

  def afterExecution(self):
    self.finishDrawing()

  def finishDrawing(self):
    # see http://stackoverflow.com/questions/7141509/tkinter-wait-for-item-in-queue/7198960#7198960
    self.root.event_generate('<<DrawingUpdate>>', when='tail')

  def updateViewport(self):
    self.canvas.update_idletasks() # required to make winfo_* return correct values
    self.pixel_viewport = (self.canvas.winfo_width(), self.canvas.winfo_height())
    self.viewport = (
      (self.pixel_viewport[0]-self.XOFFSET)/self.XMUL,
      (self.pixel_viewport[1]-self.YOFFSET)/self.YMUL)
    debug("updateViewport yielded {} / {}".format(repr(self.pixel_viewport), repr(self.viewport)));

  def clear(self):
    def clearInternal(self):
      for i in self.items:
        self.canvas.delete(i)
      self.items = []
    self.drawq.put(lambda self_: clearInternal(self_))

  def transformAsp2Pixel(self, spec, which):
    '''
    spec is either a number or 'rel(number)' or 'abs(number)'
    which is either 'x' or 'y'
    '''
    debug('transformAsp2Pixel({},{})'.format(repr(spec),which))
    assert(isinstance(spec,api.Term))
    assert(which in ['x', 'y'])
    if spec.pred == 'rel':
      return self.coordinate2canvas(str(spec.args[0]), 'rel', which)
    elif spec.pred == 'abs':
      return self.coordinate2canvas(str(spec.args[0]), 'abs', which)
    elif isinstance(spec.pred, int) or isinstance(spec.pred, str):
      return self.coordinate2canvas(spec.pred, 'abs', which)
    else:
      raise Exception("cannot interpret coordinate spec '{}'/{}".format(spec, which))
    
  def coordinate2canvas(self, spec, mode='abs', which='x'):
    assert(mode in ['abs','rel'] and which in ['x','y'])
    spec = float(spec)
    if mode == 'abs':
      if which == 'x':
        offset, mul = self.XOFFSET, self.XMUL
      else:
        offset, mul = self.YOFFSET, self.YMUL
      return offset+spec*mul
    elif mode == 'rel':
      if which == 'x':
        portsize = self.pixel_viewport[0]
      else:
        portsize = self.pixel_viewport[1]
      return portsize*spec/100.0
 
  def cachedImage(self, filename):
    def createPhotoImage(imgfilename):
      debug("createPhotoImage({})".format(repr(imgfilename)))
      return tk.PhotoImage(file=imgfilename)
    if filename not in self.imageCache:
      self.imageCache[filename] = createPhotoImage(filename)
    return self.imageCache[filename]

  def drawingUpdate(self, event):
    '''
    just call all queue elements with self
    see http://stackoverflow.com/questions/7141509/tkinter-wait-for-item-in-queue/7198960#7198960
    '''
    #warn("## drawevent "+repr(event));
    try:
      while not self.drawq.empty():
        drawevent = self.drawq.get()
        #print(repr(drawevent))
        drawevent(self)
      self.updateViewport()
    except:
      warn(traceback.format_exc())
      os._exit(-1)

class DrawRectangleAction(api.Action):
  '''
  action handler class for
    drawRectangle(x1spec,y1spec,x2spec,y2spec[,color])
  '''
  def __init__(self, window):
    self.window = window
  @property
  def interface(self):
    return ('drawRectangle', 5, 4)
  def execute(self, ainput):
    assert(isinstance(ainput, api.ActionInput))
    assert(len(ainput.inputs) in [4, 5])
    debug('DrawRectangleAction.execute({})'.format(ainput))
    x1, y1, x2, y2 = ainput.inputs[0:4]
    color = 'white'
    if len(ainput.inputs) > 4:
      color = ainput.inputs[4]
    px1 = self.window.transformAsp2Pixel(x1,'x')
    py1 = self.window.transformAsp2Pixel(y1,'y')
    px2 = self.window.transformAsp2Pixel(x2,'x')
    py2 = self.window.transformAsp2Pixel(y2,'y')
    self.window.drawq.put(lambda window:
      DrawRectangleAction.drawRectangleInternal(
        window, px1, py1, px2, py2, color))
  @staticmethod
  def drawRectangleInternal(window, x1, y1, x2, y2, color):
    '''this method runs in the TK thread'''
    debug("drawRectangleInternal {}/{} {}/{} {}".format(x1, y1, x2, y2, color))
    window.items.append(window.canvas.create_rectangle(
      x1, y1, x2, y2, fill=color))

class DrawImageAction(api.Action):
  '''
  action handler class for
    drawImage(xspec,spec,imagefilename)
  '''
  def __init__(self, window):
    self.window = window
  @property
  def interface(self):
    return ('drawImage', 3, 3)
  def execute(self, ainput):
    assert(isinstance(ainput, api.ActionInput))
    assert(len(ainput.inputs) in [3, 3])
    debug('DrawImageAction.execute({})'.format(ainput))
    x, y, filename = ainput.inputs
    px = self.window.transformAsp2Pixel(x,'x')
    py = self.window.transformAsp2Pixel(y,'y')
    f = str(filename).strip('"')
    self.window.drawq.put(lambda window:
      DrawImageAction.drawImageInternal(window, px, py, f))
  @staticmethod
  def drawImageInternal(window, x, y, filename):
    '''this method runs in the TK thread'''
    debug("drawImageInternal {}/{} {}".format(x, y, filename))
    img = window.cachedImage(filename)
    window.items.append( window.canvas.create_image(
      x, y, image=img, anchor=tk.SW) )

class DrawTextActionBase(api.Action):
  '''
  action handler base class for
    drawText(xspec,yspec,text[,color])
      draws text at specified position
    drawTextCentered(text[,color])
      draws a list from top to bottom
  '''
  def __init__(self, window):
    self.window = window
  @staticmethod
  def drawTextInternal(window, x, y, text, color, align):
    debug("drawTextInternal {}/{},{},{} {}".format(
      x, y, align, color, repr(text)))
    if align == 'center':
      anchor, justify = tk.S, tk.CENTER
    else:
      anchor, justify = tk.SW, tk.LEFT
    window.items.append( window.canvas.create_text(
      x, y, text=text, fill=color, font=window.FONT,
      anchor=anchor, justify=justify))

class DrawTextAction(DrawTextActionBase):
  '''
  action handler class for
    drawText(xspec,yspec,text[,color])
      draws text at specified position
  '''
  def __init__(self, window):
    DrawTextActionBase.__init__(self, window)
  @property
  def interface(self):
    return ('drawText', 4, 3)
  def execute(self, ainput):
    assert(isinstance(ainput, api.ActionInput))
    assert(len(ainput.inputs) in [3, 4])
    debug('DrawTextAction.execute({})'.format(ainput))
    x, y, text = ainput.inputs[0:3]
    try:
      text = text.unquoted()
    except:
      text = str(text)
    color = 'white'
    if len(ainput.inputs) > 3:
      color = ainput.inputs[3]
    px = self.window.transformAsp2Pixel(x,'x')
    py = self.window.transformAsp2Pixel(y,'y')
    self.window.drawq.put(lambda window:
      DrawTextActionBase.drawTextInternal(
        window, px, py, text, color, 'left'))

class DrawTextCenteredAction(DrawTextActionBase):
  '''
  action handler class for
    drawTextCentered(text[,color])
      draws a list from top to bottom
  '''
  def __init__(self, window):
    DrawTextActionBase.__init__(self, window)
    self.resetCursor()
  @property
  def interface(self):
    return ('drawTextCentered', 2, 1)
  def resetCursor(self):
    self.cursorY = 1
  def execute(self, ainput):
    assert(isinstance(ainput, api.ActionInput))
    assert(len(ainput.inputs) in [1, 2])
    debug('DrawTextCenteredAction.execute({})'.format(ainput))
    text = ainput.inputs[0]
    try:
      text = text.unquoted()
    except:
      text = str(text)
    color = 'white'
    if len(ainput.inputs) > 1:
      color = ainput.inputs[1]
    px = self.window.coordinate2canvas(50, 'rel', 'x')
    advance = 1 + text.count('\n')
    self.cursorY += advance
    py = self.window.coordinate2canvas(self.cursorY, 'abs', 'y')
    self.window.drawq.put(lambda window:
      DrawTextActionBase.drawTextInternal(
        window, px, py, text, color, 'center'))

class Plugin(threading.Thread,api.EventAPI,api.DoAPI):
  def __init__(self):
    threading.Thread.__init__(self)
    api.EventAPI.__init__(self)
    # do not initialize any Tk here!
    # It has to be done in thread in run()
    # (Tkinter is not threadsafe)
    self.daemon = True
    self.keyboard = None
    self.window = None
    self.root = None
    self.startupFinished = threading.Event()
    self.start()
    # wait until run() has initialized all attributes (window, ...)
    if not self.startupFinished.wait(1.0):
      raise Exception("Plugin thread did not start within timeout!")

  def setQueue(self, queue):
    if not self.keyboard:
      warn("tkinter.Plugin ignoring setQueue because startup not finished")
      return
    api.EventAPI.setQueue(self, queue)
    self.keyboard.setQueue(queue)

  def actions(self):
    if not self.window:
      warn("tkinter.Plugin ignoring actions because startup not finished")
      return
    return self.window.actions()

  def beforeExecution(self):
    self.window.beforeExecution()

  def afterExecution(self):
    self.window.afterExecution()

  def run(self):
    self.root = tk.Tk()
    self.keyboard = KeyboardHandler(self.root)
    self.window = WindowHandler(self.root)

    self.root.bind('<<Quit>>', lambda x: self.root.quit())
    #self.root.bind('<Configure>', lambda x: self.configureEvent(x)) # resize, move window

    # notify constructor that we finished initializing
    self.startupFinished.set()
    self.root.mainloop()

  def quit(self):
    self.root.event_generate('<<Quit>>', when='tail')






  def configureEvent(self, evt):
    '''
    event handler for <Configure>
    (window resize or location change)
    '''
    self.eventq.put_once(MyEvent('resize'))

  def clear(self):
    def clearInternal(self):
      for i in self.items:
        self.canvas.delete(i)
      self.items = []
    self.drawq.put(lambda self_: clearInternal(self_))


class Persistence:
  def __init__(self, fluents):
    # load from file, if existing
    self.fluents = fluents
    # build a regex that recognizes fluents, with or without arguments
    fluents_regexes = map(lambda f: "({}(\(.*\))?)".format(f), self.fluents)
    self.rfluents = re.compile('|'.join(fluents_regexes))
    self.data = { 'sit':set(), 'val':set() }
    if len(fluents) > 0:
      try:
        with open('persistence.pickle', 'rb') as f:
          p = pickle.load(f)
          if not isinstance(p, dict):
            raise Exception("found pickle in wrong format!")
          if p['fluents'] != self.fluents:
            raise Exception("found pickle with wrong fluents {}!".format(p['fluents']))
          self.data = p['data']
      except Exception as e:
        # warn but ignore
        print "Warning:", str(e)

  def get(self):
    return (self.data['sit'], self.data['val'])
    
  def store(self, sit, val):
    if len(self.fluents) > 0:
      self.data = { 'sit': sit, 'val': set() }
      # store only those fluents that should be persistent!
      for fluent, value, t in val:
        if self.rfluents.match(fluent) != None:
          self.data['val'].add( (fluent, value, t) )
        else:
          #warnVerbose("fluent {} did not match re".format(fluent))
          pass
      #warnVerbose('storing persistence {} and {}'.format(self.fluents, self.data)) 
      with open('persistence.pickle', 'wb') as f:
        to_store = { 'fluents':self.fluents, 'data':self.data }
        pickle.dump(to_store, f)




import sys
import os
import argparse
import importlib
import logging
import traceback

import clingo

from .. import core
from .common import *

ASAP_PATH = os.path.join(
  os.path.dirname(os.path.realpath(__file__)),
  '../')
CORE_PATH = os.path.join(ASAP_PATH,'./core/')

class EventCallable:
  def __init__(self, currentEvent):
    self.event = termToClingoSym(currentEvent.term)
    if __debug__:
      logging.debug("@event() = {}".format(repr(self.event)))
  def __call__(self):
    ret = [self.event]
    #debug("@event() = {}".format(repr(ret)))
    return ret

class PrevvalCallable:
  def __init__(self, fluents):
    self.fluents = fluents
    self.ret = [ (termToClingoSym(kv[0]), termToClingoSym(kv[1])) for kv in self.fluents.items() ]
    if __debug__:
      logging.debug("@prevval() = {}".format(repr(self.ret)))
  def __call__(self):
    #debug("@prevval() = {}".format(repr(ret)))
    return self.ret

class ExternalCallable:
  def __init__(self, external, environment, exceptions):
    self.external = external
    self.name, self.in_arity, self.in_arity_mand, self.out_arity = external.interface
    self.environment = environment
    self.exceptions = exceptions
  def __call__(self, *arguments):
    #debug("external {} called with {}".format(self.name, repr(arguments)))
    try:
      nargs = len(arguments)
      if nargs < self.in_arity_mand or nargs > self.in_arity:
        raise Exception("invalid number of arguments for external {}: got {}, expected between {} and {}".format(
          self.name, repr(arguments), self.in_arity_mand, self.in_arity))
      apiArguments = [ clingoSymToTerm(a) for a in arguments ]
      queryinput = api.ExternalInput(self.environment, apiArguments)
      #debug("calling external {} with {}".format(self.name, repr(apiArguments)))
      out_tuples = self.external.query(queryinput)
      #if __debug__:
      #  logging.debug("external {}{} returned {}".format(self.name, repr(apiArguments), repr(out_tuples)))
      ret = []
      for ot in out_tuples:
        if len(ot) != self.out_arity:
          logging.warning("external {} returned output {} which is not of expected arity {} (ignoring)".format(self.name, repr(ot), self.out_arity))
        if not all([isinstance(t, api.Term) for t in ot]):
          raise Exception("external {} returned {} which are not api.Term instances".format(self.name, repr(list(ot))))
        otclingo = [termToClingoSym(x) for x in ot]
        # clingo expects a fun or a tuple of more than one thing
        if len(otclingo) == 1:
          clingotuple = otclingo[0]
        else:
          clingotuple = tuple(otclingo)
        #debug("@{}({}) = {}".format(self.name, repr(arguments), repr(clingotuple)))
        ret.append(clingotuple)
      if __debug__:
        logging.debug("gringo external {}{} returns {}".format(self.name, repr(arguments), repr(ret)))
      if len(ret) == 0 and self.out_arity == 1:
        # add dummy tuple, otherwise Gringo ignores (at least in 5.2.0)
        # TODO bug report to potassco
        ret.append( clingo.Function('dummy') )
      return ret
    except Exception as e:
      logging.warning(traceback.format_exc())
      self.exceptions.append("exception in external {}: {}".format(self.name, traceback.format_exc()))

class GroundingContext:
  '''
  this is used as 'context' for calling Control.ground
  it provides external grounding functions of form @func(arg1, ...)
  '''
  def __init__(self, externals, environment, fluents, currentEvent):
    self.exceptions = []
    self.externals = externals
    self.environment = environment
    self.fluents = fluents
    self.currentEvent = currentEvent
    self.callables = dict()
    self.callables['event'] = EventCallable(self.currentEvent)
    self.callables['prevval'] = PrevvalCallable(self.fluents)
    for name, e in externals.items():
      self.callables[name] = ExternalCallable(e, self.environment, self.exceptions)

  def __getattr__(self, name):
    '''
    called if a non-existing attribute (method, member variable) is used
    '''
    if name in self.callables:
      return self.callables[name]
    # an exception raised here is printed but otherwise ignored by clingo code
    # we collect exceptions and raise them from outside
    self.exceptions.append("GroundingContext.__getattr__ for '{}' (did you forget to load an external plugin?)".format(name))

    # return Fun(str(int(delay_ms)))
    #return Fun('"' + str(a).strip('"') + str(b).strip('"') + '"')
    #return Fun('"' + str(a).strip('"') + str(b).strip('"') + str(c).strip('"') + str(d).strip('"') + '"')

class ClingoBackend(api.Backend):
  '''
  the clingo realization of ASAP
  '''
  def __init__(self):
    self.exit = None # exit code, or None if no exit requested
    self.eq = core.EventQueueImpl()
  def evaluate(self, program, plugins_=[], environment=None, fluentfile=None):
    # prepare queue
    self.eq.append(api.Event(api.Term('initial_event')))
    doApis, eventApis, externalApis = core.separateApis(plugins_)
    core.registerQueue(eventApis, self.eq)
    # get dictionaries of actions and externals
    actions, externals = core.actionsFromApis(doApis), core.externalsFromApis(externalApis)

    rewritten = core.rewrite(program, target='clasp')
    fluents = core.FluentStorage()

    def utf8fromfile(path):
      with open(path, 'r', encoding='utf8') as f:
        return f.read()
    auxcode = utf8fromfile(os.path.join(CORE_PATH, 'fluents-common.lp')) + \
      utf8fromfile(os.path.join(CORE_PATH, 'fluents-clasp.lp'))

    logging.debug("starting main loop")
    iteration = 1
    while self.exit is None:
      currentEvent = self.eq.pop()
      ctx = GroundingContext(externals, environment, fluents, currentEvent)

      clingoargs = ['2']
      if logging.getLogger().isEnabledFor(logging.DEBUG):
        # output ground program
        clingoargs.append('--verbose')
        clingoargs.append('--output-debug=all')
      else:
        clingoargs.append('--warn=no-atom-undefined')
      parts = [ ('base', []) ]

      ctrl = clingo.Control(clingoargs)
      ctrl.add('base', [], rewritten + auxcode)
      try:
        ctrl.ground(parts, ctx)
      except:
        logging.critical("ctrl.ground exception "+traceback.format_exc())
      if len(ctx.exceptions) > 0:
        raise Exception('grounding exception: '+', '.join(ctx.exceptions))
      
      answersets = []
      def onModel(mdl):
        if __debug__:
          logging.debug('onModel got '+str(mdl))
        answersets.append(set([ clingoSymToTerm(a) for a in mdl.symbols(atoms=True) ]))
      ret = ctrl.solve(on_model=onModel)

      if len(answersets) == 0:
        raise Exception("no answer sets!")
      elif len(answersets) != 1:
        raise Exception("more than one answer set!")
      answerset = answersets[0]
      if __debug__:
        logging.debug("processing answer set {}".format(answerset))

      fluents = core.extractFluents(answerset)
      if __debug__:
        logging.debug("new fluent values: {}".format(fluents.nice()))

      actionSchedule = core.extractActions(answerset)
      core.executeActionSchedule(doApis, actions, environment, answerset, actionSchedule)
      logging.debug("iteration %d finished",iteration)
      iteration += 1

class ExitPlugin(api.DoAPI):
  '''plugin providing exit() action'''

  class ExitAction(api.Action):
    def __init__(self, clingo_backend):
      self.clingo_backend = clingo_backend
    @property
    def interface(self):
      return ('exit', 1, 1)
    def execute(self, ainput):
      assert(isinstance(ainput, api.ActionInput))
      assert(len(ainput.inputs) in [1, 1])
      if __debug__:
        logging.debug('ExitAction.execute({})'.format(repr(ainput)))
      try:
        # try to interpret exit code
        code = int(ainput.inputs[0].pred)
      except:
        traceback.format_exc()
        code = -1
      self.clingo_backend.exit = code

  def __init__(self, clingo_backend):
    self.clingo_backend = clingo_backend
  def actions(self):
    return [self.ExitAction(self.clingo_backend)]

class ImmediatePlugin(api.DoAPI):
  '''plugin providing immediate(<info>) action'''

  class ImmediateAction(api.Action):
    def __init__(self, clingo_backend):
      self.clingo_backend = clingo_backend
    @property
    def interface(self):
      return ('immediate', 1, 1)
    def execute(self, ainput):
      assert(isinstance(ainput, api.ActionInput))
      assert(len(ainput.inputs) in [1, 1])
      self.clingo_backend.eq.immediate(api.Event(api.Term('immediate', [ainput.inputs[0]])))

  def __init__(self, clingo_backend):
    self.clingo_backend = clingo_backend
  def actions(self):
    return [self.ImmediateAction(self.clingo_backend)]

loadedPlugins = {}
def loadPlugins(pluginnames):
  loaded_plugins_in_order = []
  for pn in pluginnames:
    pmodule = None
    try:
      pmodule = importlib.import_module(pn)
    except:
      raise Exception("loading plugin module '{}' failed: {}".format(pn, traceback.format_exc()))
    else:
      try:
        p = pmodule.Plugin()
        loadedPlugins[pn] = p
        loaded_plugins_in_order.append(p)
      except:
        raise Exception("instantiating plugin '{}' failed: {}".format(pn, traceback.format_exc()))
  return loaded_plugins_in_order




from asap.core import api

import sys
import threading
import logging
import traceback

# we again use clingo here (independently from the main ASAP instance)
import clingo
import asap.clingo.common as clingo_common

class SolverStartAction(api.Action):
  '''
  Do.start_solver(<name>,<pred>) 2 mandatory arguments
  <name> is string
  <pred> is a binary predicate:
  * <pred>(file,<file>) adds content of <file> to the program to be evaluated
  * <pred>(rule,<rule>) adds <rule> to the program to be evaluated
  * <pred>(limit,<int>) uses <int> as a limit for the answer sets to be enumerated
  '''
  def __init__(self, plugin):
    self.plugin = plugin

  @property
  def interface(self):
    return ('solver_start', 2, 2)

  def execute(self, ainput):
    assert(len(ainput.inputs) == 2)
    name, pred = ainput.inputs
    name, pred = str(name), str(pred)
    # extract info using predicate
    #logging.debug("ainput.interpretation is "+repr(ainput.interpretation))
    # default: enumerate all answer sets
    program, limit = [], 0
    for atom in [ a for a in ainput.interpretation if a.pred == pred ]:
      #logging.debug("processing relevant predicate input "+repr(atom))
      type_, arg = str(atom.args[0]), atom.args[1]
      if type_ == 'file':
        fname = arg.unquoted()
        #logging.debug("solver '%s' loading from file '%s'", name, fname)
        with open(fname,'r') as f:
          program.append(f.read())
      elif type_ == 'rule':
        rule = arg.unquoted()
        #logging.debug("solver '%s' adding rule '%s'", name, rule)
        program.append(rule)
      elif type_ == 'limit':
        limit = int(arg)
    if name in self.plugin.solver:
      #logging.debug("deleting solver "+repr(name))
      del(self.plugin.solver[name])
    # TODO assemble program
    newsolver = Solver(name, self.plugin.queue, '\n'.join(program), limit)
    self.plugin.solver[name] = newsolver
    newsolver.start()

class SolverStartedExternal(api.External):
  '''
  output: solver handles that have been started
  '''
  def __init__(self, plugin):
    self.plugin = plugin
  @property
  def interface(self):
    # name, in_arity, in_arity_mandatory, out_arity
    return ('solver_started',0,0,1)
  def query(self, einput):
    assert(len(einput.inputs) == 0)
    return [ (api.Term(n_s[0]),)
             for n_s in self.plugin.solver.items()
             if n_s[1].is_started() ]

class SolverFinishedExternal(api.External):
  '''
  output: solver handles that have finished solving
  (if there is no result, no answer set handles will be available!)
  '''
  def __init__(self, plugin):
    self.plugin = plugin
  @property
  def interface(self):
    # name, in_arity, in_arity_mandatory, out_arity
    return ('solver_finished',0,0,1)
  def query(self, einput):
    assert(len(einput.inputs) == 0)
    return [ (api.Term(n_s[0]),)
             for n_s in self.plugin.solver.items()
             if n_s[1].is_finished() ]

class AnswersetsExternal(api.External):
  '''
  return handle to answer sets of solver (if solver has finished enumerating)
  input: solver handle (given to Do.solver_start)
  output: one handle for each answer set (integer)
  '''
  def __init__(self, plugin):
    self.plugin = plugin
  @property
  def interface(self):
    # name, in_arity, in_arity_mandatory, out_arity
    return ('solver_answersets',1,1,1)
  def query(self, einput):
    assert(len(einput.inputs) == 1)
    handle = str(einput.inputs[0])
    if handle in self.plugin.solver:
      solver = self.plugin.solver[handle]
      if solver.is_finished():
        return [ (api.Term(i),) for i in range(0,len(solver.get_answersets())) ]
    return []

class ExtensionExternal(api.External):
  '''
  return extension of given handle/answerset/predicate
  (if solver has finished enumerating)
  input: solver handle (str, given to Do.solver_start)
  input: answer set handle (int)
  input: predicate (str)
  output: composite term c(...) containing tuple of extension
  '''
  def __init__(self, plugin):
    self.plugin = plugin
  @property
  def interface(self):
    # name, in_arity, in_arity_mandatory, out_arity
    return ('solver_extension',3,3,1)
  def query(self, einput):
    assert(len(einput.inputs) == 3)
    sname, ahandle, pred = einput.inputs
    try:
      sname, ahandle, pred = str(sname), int(ahandle), str(pred)
    except:
      if str(ahandle) == "dummy":
        return [] # see engine-clingo.py to understand where "dummy" comes from
      logging.warning("got wrong input type for solver_extension {} {} {}: {}".format(
        repr(sname), repr(ahandle), repr(pred), traceback.format_exc()))
      return []
    if sname in self.plugin.solver:
      solver = self.plugin.solver[sname]
      if solver.is_finished() and ahandle in range(0, len(solver.get_answersets())):
        answerset = solver.get_answersets()[ahandle]
        return [ (api.Term('c', t.args),) for t in answerset if t.pred == pred ]
    return []

class Solver:
  def __init__(self, name, eventqueue, program, answersetlimit=1):
    self.name = name
    self.queue = eventqueue
    self.program = program
    self.answersetlimit = answersetlimit
    self.thread = None
    self.answersets = None
  def start(self):
    assert(self.thread is None)
    self.thread = threading.Thread(target=self.run)
    self.thread.daemon = True
    self.thread.start()
  def run(self):
    # main entry point, started when solver is started
    logging.debug("solver '%s' starting: sending program", self.name)
    clingoargs = []
    #clingoargs.append('--output-debug=all')
    clingoargs.append(self.answersetlimit)
    try:
      ctrl = clingo.Control(clingoargs)
      if __debug__:
        for r in self.program.split('\n'):
          logging.debug("solver '%s' program %s", self.name, r.strip('\n'))
      ctrl.add('base', [], self.program)
      parts = [ ('base',[]) ] 
      ctrl.ground(parts)
    except:
      logging.critical("solver '%s' grounding exception %s", self.name, traceback.format_exc())
      if __debug__:
        logging.critical("aborting from solver to facilitate debugging")
        sys.exit(-1)
    logging.debug("solver '%s' grounding finished: solving", self.name)
    answersets = set()
    def onModel(mdl):
      logging.debug("solver '%s' got model %s", self.name, str(mdl))
      answersets.add(frozenset([ clingo_common.clingoSymToTerm(a) for a in mdl.symbols(atoms=True) ]))
    ret = ctrl.solve(on_model=onModel)
    #logging.debug("solver '%s' got %d answer sets", self.name, len(answersets))
    # it was a set of answer sets, but make it a list so that we can index it
    self.answersets = list(answersets)
    # TODO cleanup program (and ctrl?)
    # notify the framework with an event that this solver has finished
    self.queue.append(api.Event(api.Term('solver_done',[api.Term(self.name),])))
  def is_started(self):
    return self.thread is not None
  def is_running(self):
    return self.thread is not None and self.answersets is None
  def is_finished(self):
    return self.thread is not None and self.answersets is not None
  def get_answersets(self):
    """
    returns a list of answer sets (frozenset of api.Term)
    """
    return self.answersets

class Plugin(api.DoAPI, api.EventAPI, api.ExternalAPI):
  def __init__(self):
    api.EventAPI.__init__(self)
    # started solver instances
    # key = handle
    # value = Solver instance
    self.solver = {}

  def actions(self):
    return [SolverStartAction(self)]

  def externals(self):
    return [AnswersetsExternal(self), ExtensionExternal(self), SolverStartedExternal(self), SolverFinishedExternal(self)]





class ASPProcessor(EventThread):
  rPointXY = re.compile(r'point\(([0-9.]+),([0-9.]+)\)')
  def __init__(self, **kwargs):
    self.frontend = kwargs['frontend']
    del(kwargs['frontend'])
    self.sugarout = kwargs['sugarout']
    del(kwargs['sugarout'])
    EventThread.__init__(self, **kwargs)
    self.loadBaseProgram()

  def loadBaseProgram(self):
    # get special directives
    # #error #persistent #actions
    # write temporary file with the rest of the input
    hexfile = tempfile.NamedTemporaryFile(delete=False)
    rSpecial = re.compile(
      r'#(error|persistent|actions)\s(.*)\.')
    self.persistent = set()
    self.actions = set()
    errors = []
    try:
      with open(self.sugarout, 'rt') as f:
        for line in f:
          m = rSpecial.match(line)
          if m:
            type_, content = m.groups()
            if type_ == 'persistent':
              for item in content.split(','):
                self.persistent.add(item.strip())
            elif type_ == 'actions':
              for item in content.split(','):
                self.actions.add(item.strip())
            elif type_ == 'error':
              errors.append(content)
            else:
              assert(False and "we should handle all types above")
          else:
            hexfile.write(line)
      hexfile.close()
      if len(errors) > 0:
        raise Exception("errors in sugarizer! ({})".format(
          ', '.join(errors)))
      # load the HEX code
      edb, idb = dlvhex.loadSubprogram(hexfile.name)
      self.baseProgram = [edb, idb]
    finally:
      #print "keeping temporary hex input file " + hexfile.name
      os.unlink(hexfile.name)
      pass

    #print "Evaluating the program:" 
    #print dlvhex.getValue(self.baseProgram[1]) 
    #print "Evaluating wrt the following facts:" 
    #print dlvhex.getValue(self.baseProgramcompleteprog[0]) 

  def selectAnswerset(self, answersets):
    if len(answersets) == 0:
      raise RuntimeError("no answersets!")
    if len(answersets) != 1:
      warn("more than one answerset!")
      asSetOfStrings = map(lambda a: set(map(dlvhex.getValue, a)), answersets)
      for idx, answerset in enumerate(asSetOfStrings):
        if idx == 0:
          continue
        for idxPrev, answersetPrev in enumerate(asSetOfStrings):
          if idxPrev >= idx:
            break
          plus = answerset - answersetPrev
          minus = answersetPrev - answerset
          #warn('plus {}'.format(plus))
          #warn('minus {}'.format(minus))
          diff = map(lambda p: p+'+', plus) + map(lambda m: m+'-', minus)
          warn("#{} vs #{}: {}".format(idx, idxPrev, ', '.join(sorted(diff))))
      for idx, answerset in enumerate(answersets):
        warn("Full Answer #{}: {}\n".format(idx, dlvhex.getValue(answerset)))
      raise RuntimeError("more than one answerset!")
    return answersets[0]

  def checkForASPErrors(self, answerset):
    # check for errors
    error = False
    for atom in answerset:
      atuple = dlvhex.getTupleValues(atom)
      if len(atuple) == 2 and atuple[0] == 'error':
        warn("ASP ERROR: {}".format(repr(atuple[1])))
        error = True
    if error:
      raise RuntimeError("encountered ASP error(s)!")

  def showPredicates(self, answerset, preds):
    # this function is for debugging
    predcollection = {}
    for atom in answerset:
      atuple = dlvhex.getTupleValues(atom)
      pred = atuple[0]
      if pred in preds:
        if pred in predcollection:
          predcollection[pred].append( atuple[1:] )
        else:
          predcollection[pred] = [ atuple[1:] ]
    for pred, extension in predcollection.iteritems():
      #warn(repr(extension))
      warn("{}: {}".format(pred, ", ".join(map(lambda ext: "/".join(ext), extension))))

  def enqueueSoftEvents(self, actions):
    '''
    find soft events and execute them
    return unhandled actions
    '''
    ret = []
    for a in actions:
      aname, aargs, ser = a[0], a[1:-1], a[-1]
      if aname == 'softEvent':
        name = aargs[0]
        self.queue.put_front(MyEvent('softevent', name))
      else:
        ret.append(a)
    return ret





from asap.core import api

import sys
import random
import re
import logging

def strIfRequired(inp):
  '''
  convert an input to a string (if it is not yet)
  (warn if conversion was necessary)
  '''
  if isinstance(inp,str):
    return inp
  else:
    if __debug__:
      logging.warning("converted non-string input '{}'".format(repr(inp)))
    return str(inp)

def stripQuotes(inputs):
  '''
  remove outside quotes for all inputs
  return True, inputs' if anything was stripped,
  return False, inputs' otherwise
  '''
  changed = False
  out = []
  for i in inputs:
    i = strIfRequired(i)
    if i[0] == '"' and i[-1] == '"':
      changed = True
      out.append(i[1:-1])
    else:
      out.append(i)
  return changed, out

def conditionalQuote(quote, inp):
  if quote:
    return '"'+inp+'"'
  else:
    return inp

rCident = re.compile(r'[a-z_][A-Z0-9a-z_]*')
def quoteIfRequired(inp):
  if inp[0] != '"' and not rCident.match(inp):
    return '"'+inp+'"'
  else:
    return inp

class ChopExternal(api.External):
  '''
  remove last character of input string
  input: string
  output: string or nothing (if input empty)
  '''
  @property
  def interface(self):
    # one input, one output
    return ('chop',1,1,1)
  def query(self, einput):
    assert(len(einput.inputs) == 1)
    quoted, instrs = stripQuotes(einput.inputs)
    instr = instrs[0]
    if len(instr) == 0:
      return []
    else:
      out = conditionalQuote(quoted, instr[:-1])
      return [(api.Term(out),)]

class ConcatExternal(api.External):
  '''
  concatenate inputs
  output quoted string if one input was quoted string
  output term if no input was quoted string
  '''
  @property
  def interface(self):
    # at least 2 inputs, one output
    return ('concat',99,2,1)
  def query(self, einput):
    if __debug__:
      logging.debug("concat called with {}".format(repr(einput.inputs)))
    quoted, inputs = stripQuotes([str(i) for i in einput.inputs])
    if __debug__:
      logging.debug("concat called with {} {}".format(repr(quoted),repr(inputs)))
    out = conditionalQuote(quoted, ''.join(inputs))
    return [ (api.Term(out),) ]

#TODO make concat/chop always return strings

class RandomExternal(api.External):
  '''
  return random integer within given bounds
  input: lower, upper bound (int)
  output: random integer within bounds
  '''
  @property
  def interface(self):
    # 2 inputs, 1 output
    return ('random',2,2,1)
  def query(self, einput):
    assert(len(einput.inputs) == 2)
    #assert(isinstance(einput.inputs[0], int) and isinstance(einput.inputs[1], int))
    # convert and if this conversion fails we will get an exception anyways
    lower, upper = int(einput.inputs[0]), int(einput.inputs[1])
    out = random.randint(lower, upper)
    return [(api.Term(out),)]

class Plugin(api.ExternalAPI):
  def externals(self):
    return [ChopExternal(), ConcatExternal(), RandomExternal()]




from asap.core import api

import sys
import threading

def debug(msg):
  sys.stderr.write("D:"+msg+"\n")

class StartTimerAction(api.Action):
  '''
  Do.startTimer(name,delay) 2 mandatory arguments
  name is string
  delay is integer (milliseconds)
  '''
  def __init__(self, queue):
    self.queue = queue

  @property
  def interface(self):
    return ('startTimer', 2, 2)

  def execute(self, ainput):
    assert(len(ainput.inputs) == 2)
    name, millis = ainput.inputs
    millis = float(int(millis))
    class EventCreator:
      def __init__(self, queue, eventdata):
        self.queue = queue
        self.eventdata = eventdata
      def __call__(self):
        # enqueue event
        self.queue.append(api.Event(self.eventdata))
    debug("startTimer {} for {} ms".format(name, millis))
    t = threading.Timer(millis/1000.0, EventCreator(self.queue, api.Term('timer',[name])))
    t.start()

class Plugin(api.DoAPI, api.EventAPI):
  def __init__(self):
    api.EventAPI.__init__(self)

  def actions(self):
    if not self.queue:
      raise Exception('timer.Plugin.actions() can only be called after provding the queue via timer.Plugin.setQueue()')
    return [StartTimerAction(self.queue)]




# This pyclingo installation module was taken from
# "HEXLite Python-based solver for a fragment of HEX"
# Copyright (C) 2017  Peter Schueller <schueller.p@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# this module supports building clingo module from the distribution

import os, sys, subprocess, logging, tempfile, traceback

def msg(m):
  sys.stderr.write(m+'\n')

class InstallerBase:
  # you can change this to reduce/increase number of parallel jobs
  MAKEARGS = ['VERBOSE=1', '--jobs=4']

  def __init__(self):
    self.allyes = False
    self.tmpdir = None
    self.INSTALLDIR = os.path.expanduser('~/.asap/')

  def run_cmd(self, cmd, **args):
    logging.info("running command '{}'".format(' '.join(cmd)))
    subprocess.check_call(cmd, **args)

  def prompt_user(self, message):
    msg(message)
    if self.allyes:
      return True
    answer = None
    while answer not in ['n', 'y', 's', 'a']:
      msg("Continue? (y/n/s/a) (yes, no, skip one, all yes)")
      answer = sys.stdin.readline().strip()
    if answer == 'n':
      raise Exception("User aborted setup")
    if answer == 'a':
      self.allyes = True
    if answer in ['y','a']:
      return True
    # skip
    return False

  def ensurepackages(self, packages):
    cmd = ['dpkg-query', '-W', "-f=${Package}\\n"]
    logging.info('obtaining list of installed packages with command '+repr(cmd))
    allpackages = subprocess.check_output(cmd).decode('utf8')
    #logging.debug('got list '+repr(allpackages))
    allpackages = set([pkg.strip() for pkg in allpackages.split('\n')])
    need = [pkg for pkg in packages if pkg not in allpackages]
    if len(need) > 0:
      logging.info("did not find required packages {} via dpkg".format(repr(need)))
      prompt = "Will run 'sudo apt-get install {}' (you may do this yourself and restart this script, no other installation part requires sudo)".format(repr(need))
      if not self.prompt_user(prompt):
        return
      subprocess.check_call(['sudo', 'apt-get', 'install']+list(need))

  def maketargetdir(self):
    d = self.INSTALLDIR
    if not os.path.isdir(d):
      try:
        prompt = "Will create directory {} for installation".format(d)
        if not self.prompt_user(prompt):
          return
        os.makedirs(d)
      except:
        raise Exception('could not create output directory '+d)

  def prepare(self, packages):
    self.ensurepackages(packages)
    self.maketargetdir()
    self.tmpdir = tempfile.TemporaryDirectory()

  def build_install(self, srcdir):
    cmd = [
      'cmake', '.', '-DCMAKE_BUILD_TYPE=Release', '-DCMAKE_INSTALL_PREFIX='+self.INSTALLDIR,
      '-DCLINGO_BUILD_PY_SHARED=ON', '-DPYCLINGO_INSTALL_DIR='+self.INSTALLDIR,
      '-DPYTHON_EXECUTABLE=/usr/bin/python3']
    prompt = "Will next run cmake in {} with command {}".format(srcdir, repr(cmd))
    if self.prompt_user(prompt):
      self.run_cmd(cmd, cwd=srcdir)
    prompt = "Will next run make and make install with arguments MAKEARGS={}".format(repr(self.MAKEARGS))
    if self.prompt_user(prompt):
      self.run_cmd(['make']+self.MAKEARGS, cwd=srcdir)
      self.run_cmd(['make', 'install'], cwd=srcdir)
      logging.info("removing temporary directory")
      self.tmpdir.cleanup()

class ArchivedReleaseInstaller(InstallerBase):
  ARCHIVE = 'v5.2.0.tar.gz'
  URL = 'https://github.com/potassco/clingo/archive/'+ARCHIVE
  DIR_IN_ARCHIVE = 'clingo-5.2.0'

  def __init__(self):
    InstallerBase.__init__(self)

  def download(self):
    logging.info("Dowloading " + self.URL)
    self.run_cmd(['wget', self.URL, '--output-document='+os.path.join(self.tmpdir.name, self.ARCHIVE)])

  def unpack(self):
    logging.info("Unpacking downloaded archive to " + self.tmpdir.name)
    self.run_cmd(['tar', 'xzf', self.ARCHIVE], cwd=self.tmpdir.name)
    self.SRCDIR = os.path.join(self.tmpdir.name, self.DIR_IN_ARCHIVE)

  def doit(self, packages):
    self.prepare(packages)
    self.download()
    self.unpack()
    self.build_install(self.SRCDIR)

class GitCheckoutInstaller(InstallerBase):
  GITURL = 'https://github.com/potassco/clingo.git'
  VERSION = '58ee38'

  def __init__(self):
    InstallerBase.__init__(self)

  def clone(self):
    logging.info("Cloning {} into {}".format(self.GITURL, self.tmpdir.name))
    self.run_cmd(['git', 'clone', '--recursive', self.GITURL, os.path.join(self.tmpdir.name)])

  def checkout(self):
    logging.info("Checking out {}".format(self.VERSION))
    self.run_cmd(['git', 'checkout', self.VERSION], cwd=self.tmpdir.name)
    self.run_cmd(['git', 'submodule', 'update', '--recursive'], cwd=self.tmpdir.name)

  def doit(self, packages):
    self.prepare(packages)
    self.clone()
    self.checkout()
    self.build_install(self.tmpdir.name)

def build():
  try:
    lsbid = subprocess.check_output(['lsb_release', '--short', '--id']).decode('utf8').strip()
    lsbrelease = subprocess.check_output(['lsb_release', '--short', '--release']).decode('utf8').strip()
    logging.debug('got LSB id {} and release {}'.format(lsbid, lsbrelease))
    # for installing from release packages
    #USUALPACKAGES = ['wget', 'tar', 'gzip', 'cmake', 'g++', 'libpython3-dev']
    # for installing from git
    USUALPACKAGES = ['wget', 'tar', 'gzip', 'cmake', 'g++', 'libpython3-dev', 'bison', 're2c', 'git']
    # for the java-based rewriter on debian-flavor OS
    USUALPACKAGES.append('default-jre')
    # for tkinter on debian-flavor OS
    USUALPACKAGES.append('python3-tk')
    IMPOSSIBLE = {
      ('Ubuntu', '14.04'): 'Does not contain sufficiently new cmake and g++ for automatic build (you can try to manually install these)'
    }
    UBUNTU_TESTED = ['16.04', '16.10', '17.04']
    DEBIAN_TESTED = ['?']
    inst = ArchivedReleaseInstaller()
    inst = GitCheckoutInstaller()
    if lsbid == 'Ubuntu':
      if lsbrelease == '14.04':
        raise Exception("Ubuntu 14.04 does not contain modern cmake required for building clingo")
      elif lsbrelease in UBUNTU_TESTED:
        logging.info('installing for tested Ubuntu version')
        inst.doit(USUALPACKAGES)
      else:
        logging.info('installing for untested Ubuntu version {} (tested = {})'.format(lsbrelease, repr(UBUNTU_TESTED)))
        inst.doit(USUALPACKAGES)
    elif lsbid == 'Debian':
      if lsbrelease in DEBIAN_TESTED:
        logging.info('installing for tested Debian version')
        inst.doit(USUALPACKAGES)
      else:
        logging.info('installing for untested Debian version {} (tested = {})'.format(lsbrelease, repr(DEBIAN_TESTED)))
        inst.doit(USUALPACKAGES)
    else:
      logging.info('installing for untested Linux id {} and release {} (tested = Ubuntu {} and Debian {})'.format(
        repr(lsbid), repr(lsbrelease),
        repr(UBUNTU_TESTED), repr(DEBIAN_TESTED)))
      inst.doit(USUALPACKAGES)
  except IOError:
    logging.critical("We are sorry, could not determine your operating system." +
      " Please contact the developers.")
    return False
  except:
    logging.critical("Unexpected Exception:"+traceback.format_exc()+"\nPlease contact the developers.")
    return False
  return True




from asap.core import api

# this module is provided by clingo.so built using `scons pyclingo` in the potassco suite
import clingo

def termToClingoSym(term):
  '''
  convert Term to clingo.Function
  '''
  #debug("converting term '{}' to clingo.Function".format(repr(term)))
  if not isinstance(term, api.Term):
    raise Exception("termToClingoSym expected api.Term but got '{}'".format(repr(term)))
  ret = None
  if len(term.args) == 0:
    if isinstance(term.pred, int):
      ret = clingo.Number(term.pred)
    elif term.pred[0] == '"':
      ret = clingo.String(term.pred[1:-1])
    else:
      ret = clingo.Function(term.pred)
  else:
    ret = clingo.Function(term.pred, [termToClingoSym(x) for x in term.args])
  #debug("converted term '{}' to clingo.Function '{}'".format(repr(term), repr(ret)))
  return ret


def clingoSymToTerm(sym):
  '''
  convert clingo.Function to Term
  '''
  #debug("converting clingo.Function '{}' to Term".format(repr(fun)))
  assert(isinstance(sym, clingo.Symbol))
  if sym.type == clingo.SymbolType.Number:
    ret = api.Term(int(sym.number))
  elif sym.type == clingo.SymbolType.String:
    ret = api.Term('"'+sym.string+'"')
  elif sym.type == clingo.SymbolType.Function:
    ret = api.Term(sym.name, [clingoSymToTerm(x) for x in sym.arguments])
  else:
    raise Exception("cannot convert '{}' to Term!".format(repr(sym)))
  #debug("converted clingo.Function '{}' to Term '{}'".format(repr(sym), repr(ret)))
  return ret



#!/usr/bin/env python3

import sys
import platform
import setuptools

if not platform.python_version().startswith('3'):
  sys.stdout.write("Please use Python 3!\n")
  sys.exit(-1)

def readme():
  with open('README', 'wb') as of:
    with open('README.rst', 'rb') as i:
      out = b''.join([ line for line in i if not line.startswith(b'[')])
      of.write(out)
      return out.decode('utf8')

setuptools.setup(name='asap',
      version='2.0.0a6',
      description='ASAP Answer Set Application Programming',
      long_description=readme(),
      classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Utilities',
      ],
      #url='https://github.com/peschue/asap',
      url='https://bitbucket.org/peterschueller/asap-aspetris',
      author='Peter Schuller, Antonius Weinzierl',
      author_email='schueller.p@gmail.com',
      license='BSD',
      packages=['asap', 'asap.core', 'asap.clingo', 'asap.plugins', 'asap.examples', 'asap_tetris', 'asap_aplagent'],
      package_data = {
        'asap.core' : ['*.jar'],
        'asap.examples' : ['*.asap'],
        'asap-tetris' : ['*.asap', 'resources/*.png'],
        'asap-aplagent' : ['*.asap', '*.apl'],
      },
      scripts=[
        'bin/asap',
        'bin/asap-aplagent',
        'bin/asap-tetris',
        'bin/asap-example-onoffswitch',
      ],
      zip_safe=False)
